from __future__ import absolute_import, unicode_literals

import json
import logging
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from urlobject import URLObject as URL

from .exceptions import ChinupCanceled, PagingError
from .lowlevel import parse_fb_exception
from .queue import get_queue
from .util import partition
from . import settings


logger = logging.getLogger(__name__)


class Chinup(object):
    """
    A single FB request/response. This shouldn't be instantiated directly,
    rather the caller should use a ChinupBar:

        chinup = ChinupBar(token='XYZ').get('me')

    This returns a chinup which is a lazy request. It's on the queue and will
    be processed when convenient. The chinup can be access as follows:

        chinup.response = raw response from FB
        chinup.data = dict or list from FB, depending on endpoint
        chinup[key] = shortcut for chinup.data[key]
        key in chinup = shortcut for key in chinup.data

    The preferred method for accessing a list response is to iterate or listify
    the chinup directly. This will automatically advance through paged data,
    whereas accessing chinup.data will not.

        list(chinup)
            OR
        for d in chinup:
            do something clever with d
    """

    def __init__(self, queue, method, path, data, token=None,
                 raise_exceptions=True, callback=None, prefetch_next_page=True):
        self.queue = queue
        self.request = dict(method=method, path=path, data=data)
        self.token = token
        self.raise_exceptions = raise_exceptions
        self.callback = callback
        self.prefetch_next_page = prefetch_next_page
        self._response = None
        self._exception = None
        self._next_page = None
        queue.append(self)

    def __unicode__(self, extra=''):
        r = self._response
        if (isinstance(r, dict) and r.get('code') == 200 and 'headers' in r and
                not settings.DEBUG_HEADERS):
            r = dict(r)
            del r['headers']
        return ('{0.request[method]} {0.request[path]} '
                '{2}data={0.request[data]} response={1!r}').format(
                    self, r, extra + ' ' if extra else '')

    def __repr__(self):
        return '<{0.__class__.__name__} id={1} {0} >'.format(self, id(self))

    @property
    def completed(self):
        return any(x is not None for x in [self._response, self._exception])

    def _sync(self):
        if not self.completed:
            self.queue.sync(self)

    @property
    def response(self):
        self._sync()
        return self._response

    @response.setter
    def response(self, response):
        self._response = response

        # Decode and promote response body. The body can be None if the HTTP
        # status code isn't 200, for example 302 with a Location header.
        if isinstance(response, dict) and response.get('body') is not None:
            try:
                body = json.loads(response['body'])
            except ValueError as e:
                if not self._exception:
                    self._exception = e
            else:
                if (isinstance(body, dict) and ('data' in body or
                                                'error' in body)):
                    response.update(body)
                else:
                    response['data'] = body

                # Response body successfully decoded and promoted;
                # remove it from the dict to reduce confusion in logging etc.
                del response['body']

        # Keep parity with facepy exceptions.
        if not self._exception:
            self._exception = parse_fb_exception(response)

        # If this chinup has an associated callback, call it now. This allows
        # the caller to chain chinups, for example an async ads report. Don't
        # use this if you're not sure you need it.
        if self.callback:
            self.callback(self)

        # Prepare to fetch the next page.
        if self.prefetch_next_page:
            self.fetch_next_page()

    def _response_get(self, name):
        self._maybe_raise_exception()
        if isinstance(self.response, dict):
            return self.response.get(name)

    @property
    def data(self):
        return self._response_get('data')

    @property
    def error(self):
        return self._response_get('error')

    def cancel(self):
        self.exception = ChinupCanceled()

    @property
    def exception(self):
        self._sync()
        return self._exception

    @exception.setter
    def exception(self, value):
        self._exception = value

    def _maybe_raise_exception(self):
        if self.raise_exceptions and self.exception:
            logger.warning("Raising %s for %r", self.exception.__class__.__name__, self)
            raise self.exception

    def fetch_next_page(self):
        """
        Prepare to load the next page by putting a chinup on the queue.
        This doesn't actually do anything, of course, until .data or similar is
        accessed.
        """
        if self._next_page:
            return

        if not isinstance(self.response, dict):
            return

        try:
            next_link = self.response['paging']['next']
        except KeyError:
            return

        # FB provides a completely bogus "next" link to the insights call.
        if '/server.php' in next_link:
            return

        # FB can provide a "next" link when there's obviously nothing more.
        # Even worse, the "next" link on the adreportstats endpoint results in
        # a 500 error if you request past the end. Try to avoid that.
        limit = (self.response.get('limit') or
                 URL(next_link).query_dict.get('limit'))
        if limit and len(self.data) < int(limit):
            return

        # Putting this on the queue now enables
        # paging of one chinup to simultaneously prefetch paged data
        # for all chinups in the same queue.
        self._next_page = self._get_next_page(next_link)

    def _get_next_page(self, next_link, **kwargs):
        """
        Returns the chinup corresponding to the next page.
        This accepts kwargs for the sake of subclasses.
        """
        # Instantiate without token, since link already has it.
        return self.__class__(
            queue=self.queue,
            method=self.request['method'],
            path=URL(next_link).with_scheme('').with_netloc('')[1:],
            data=None,  # all params are in next_link
            raise_exceptions=self.raise_exceptions,
            callback=self.callback,
            prefetch_next_page=self.prefetch_next_page,
            **kwargs)

    def next_page(self):
        """
        Returns the chinup corresponding to the next page, or None if
        il n'y en a pas.
        """
        self.fetch_next_page()
        return self._next_page

    def __iter__(self):
        """
        Yields successive records from self.data, advancing through paged data
        automatically.
        """
        chinup = self
        while chinup:  # will be None on last page
            if not isinstance(chinup.data, list):
                if not self.exception:
                    self.exception = PagingError("Unexpected chinup.data while paging")
                    self.exception.chinup = chinup
                    self._maybe_raise_exception()
                break
            for d in chinup.data:
                yield d
            chinup = chinup.next_page()

    def __len__(self):
        """
        Returns the length of data in this chinup, not all pages,
        to avoid infinite recursion because Python calls len(obj) as a first
        step to list(obj).
        """
        # If self.raise_exceptions is False, then self.data might be None.
        # Don't accidentally raise an exception here with len(None).
        return len(self.data or [])

    def __nonzero__(self):
        # Prevent truth value testing from calling len(self).
        return True

    def __getitem__(self, name):
        if isinstance(name, int) and isinstance(self.data, list):
            if name < len(self.data):
                return self.data[name]
            else:
                # Invoke paging
                return list(self)[name]
        return self.data[name]

    def get(self, name, default=None):
        return self.data.get(name, default)

    def __contains__(self, name):
        assert isinstance(self.data, dict)
        return name in self.data

    def items(self):
        return self.data.items()

    def __getstate__(self):
        assert not self.callback, "can't pickle chinup with callback"
        return dict(self.__dict__,
                    queue=self.queue.app_token)

    def __setstate__(self, d):
        self.__dict__.update(d)
        if not self.completed:
            self.queue = get_queue(self.queue)
            self.queue.append(self)

    def make_request_dict(self):
        """
        Returns a dict suitable for a single request in a batch.
        """
        method = self.request['method']
        relative_url = URL(self.request['path'])
        data = self.request['data'] or {}

        if method == 'DEBUG_TOKEN':
            # This is a special case where access_token should NOT be set on
            # the relative_url, but should be passed as input_token instead.
            assert self.token, "can't debug_token without a token"
            method = 'GET'
            relative_url = URL('debug_token').set_query_params(
                input_token=self.token)

        elif self.token:
            relative_url = relative_url.set_query_params(
                access_token=self.token)

        if method != 'POST':
            # HACK since URLObject doesn't like ints in query values.
            data = {k: str(v) if isinstance(v, int) else v
                    for k, v in data.items()}
            relative_url = relative_url.set_query_params(**data)

        req = dict(
            method=method,
            relative_url=relative_url,
        )

        if method == 'POST':
            data, files = map(dict, partition(lambda d: hasattr(d[1], 'read'),
                                              data.items()))
            if data:
                req['body'] = urlencode(data)
            if files:
                req['files'] = files

        return req

    @classmethod
    def prepare_batch(cls, chinups):
        """
        Returns a tuple of (chinups, requests) where requests is a list of
        dicts appropriate for a batch request.
        """
        # Build request dicts for the first 50 chinups, limit imposed by the
        # Facebook API.
        requests = [c.make_request_dict() for c in chinups[:50]]

        # Return the full list of chinups and the possibly shorter list of
        # requests.  Note the requests still match one-to-one with the chinups
        # though, and that's important.
        return chinups, requests


class ChinupBar(object):
    chinup_class = Chinup

    def __init__(self, token=None, app_token=None,
                 raise_exceptions=True, prefetch_next_page=True):
        if not app_token:
            app_token = settings.APP_TOKEN
            if not app_token:
                raise ValueError("Either app_token or settings.APP_TOKEN is required.")

        self.app_token = app_token
        self.token = token
        self.raise_exceptions = raise_exceptions
        self.prefetch_next_page = prefetch_next_page

    def _get_queue(self):
        return get_queue(self.app_token)

    def _get_chinup(self, **kwargs):
        return self.chinup_class(**kwargs)

    def _query(self, method, path, data, defer):
        queue = self._get_queue()
        chinup = self._get_chinup(queue=queue, token=self.token,
                                  method=method, path=path, data=data,
                                  raise_exceptions=self.raise_exceptions,
                                  prefetch_next_page=self.prefetch_next_page)
        if not defer:
            queue.sync(chinup)
            # For non-deferred requests, raise exception immediately rather
            # than waiting for .data to be accessed. This is especially for
            # POST which might not check its response.
            if self.raise_exceptions and chinup.exception:
                raise chinup.exception
        return chinup

    def get(self, path, data=None, defer=True):
        return self._query('GET', path, data, defer)

    def post(self, path, data, defer=False):
        return self._query('POST', path, data, defer)

    def put(self, path, data, defer=False):
        return self._query('PUT', path, data, defer)

    def delete(self, path, data=None, defer=False):
        return self._query('DELETE', path, data, defer)

    def debug_token(self, path='', data=None, defer=True):
        return self._query('DEBUG_TOKEN', path, data, defer)


__all__ = ['Chinup', 'ChinupBar']



from __future__ import absolute_import, unicode_literals


def partition(cond, seq, parts=2):
    """
    Partition function from Erik Wilsher on Ned's blog at
    http://nedbatchelder.com/blog/200607/partition_for_python.html
    but with cond first, to match filter, map, etc.
    """
    res = [[] for i in range(parts)]
    for e in seq:
        res[int(cond(e))].append(e)
    return res



from __future__ import absolute_import, unicode_literals

from allauth.socialaccount.models import SocialToken
from django.conf import settings as django_settings
from django.contrib.auth import get_user_model
from django.db.models import Q

from . import chinup, exceptions


class NoSuchUser(exceptions.ChinupError):
    pass

exceptions.NoSuchUser = NoSuchUser


class MissingToken(exceptions.ChinupError):
    pass

exceptions.MissingToken = MissingToken


class Chinup(chinup.Chinup):

    def __init__(self, user=None, **kwargs):
        super(Chinup, self).__init__(**kwargs)
        self.user = user

    def __unicode__(self, extra=''):
        extra = '{}user={}'.format(extra and extra + ' ', self.user)
        return super(Chinup, self).__unicode__(extra=extra)

    def __getstate__(self):
        return dict(super(Chinup, self).__getstate__(),
                    user=getattr(self.user, 'pk', self.user))

    @classmethod
    def prepare_batch(cls, chinups):
        # Populate user tokens into chinups. This also immediately "completes"
        # any chinups which require a token that isn't available, by setting
        # chinup.exception.
        cls._fetch_users(chinups)
        cls._fetch_user_tokens(chinups)

        # Weed out any chinups that didn't pass token stage.
        chinups = [c for c in chinups if not c.completed]

        return super(Chinup, cls).prepare_batch(chinups)

    @classmethod
    def _fetch_users(cls, chinups):
        chinups = [c for c in chinups if not c.completed and not c.token
                   and isinstance(c.user, basestring)]
        if chinups:
            users = cls._users_dict(chinups)
            for c in chinups:
                user = users.get(c.user)
                if user:
                    c.user = user
                else:
                    c.exception = NoSuchUser("No user %r" % c.user)

    @classmethod
    def _users_dict(cls, chinups):
        User = get_user_model()
        db_users = User.objects.filter(
            Q(pk__in=set(c.user for c in chinups if isinstance(c.user, int))) |
            Q(username__in=set(c.user for c in chinups if isinstance(c.user, basestring))))
        users = {u.pk: u for u in db_users}
        users.update({u.username: u for u in db_users})
        return users

    @classmethod
    def _fetch_user_tokens(cls, chinups):
        chinups = [c for c in chinups if not c.completed and not c.token
                   and c.user]
        if chinups:
            social_tokens = cls._social_token_queryset(chinups)
            social_tokens = social_tokens.select_related('account',
                                                         'account__user')
            assert (len(set(st.account.user_id for st in social_tokens)) ==
                    len(social_tokens))
            tokens = {st.account.user_id: st.token for st in social_tokens}

            for c in chinups:
                token = tokens.get(c.user.pk)
                if token:
                    c.token = token
                else:
                    c.exception = MissingToken("No token for %r" % c.user)

    @classmethod
    def _social_token_queryset(cls, chinups, **kwargs):
        site_id = getattr(django_settings, 'SITE_ID', None)
        if site_id:
            kwargs.setdefault('app__sites__id', site_id)
        return SocialToken.objects.filter(
            account__user__in=set(c.user for c in chinups),
            **kwargs)


class ChinupBar(chinup.ChinupBar):
    chinup_class = Chinup

    def __init__(self, user=None, **kwargs):
        super(ChinupBar, self).__init__(**kwargs)
        self.user = user

    def _get_chinup(self, **kwargs):
        return super(ChinupBar, self)._get_chinup(
            user=self.user, **kwargs)


__all__ = ['Chinup', 'ChinupBar', 'NoSuchUser', 'MissingToken']



from __future__ import absolute_import, unicode_literals

import logging

from .lowlevel import batches
from . import queue, settings


logger = logging.getLogger(__name__)


class ChinupMiddleware(object):
    def process_request(self, request):
        queue.delete_queues()

    def process_response(self, request, response):
        if settings.DEBUG and batches:
            logger.info("%d requests in %d batches",
                        sum(len(b) for b in batches),
                        len(batches))
            batches[:] = []

        return response



from __future__ import absolute_import, unicode_literals

import logging
import threading

from .lowlevel import batch_request
from . import settings


logger = logging.getLogger(__name__)


_threadlocals = threading.local()


def get_queue(app_token):
    try:
        qs = _threadlocals.chinup_queues
    except AttributeError:
        qs = _threadlocals.chinup_queues = {}
    try:
        q = qs[app_token]
    except KeyError:
        q = qs[app_token] = ChinupQueue(app_token)
    return q


def reset_queue(app_token):
    try:
        del _threadlocals.chinup_queues[app_token]
    except (AttributeError, KeyError):
        pass


def delete_queues():
    try:
        del _threadlocals.chinup_queues
    except AttributeError:
        pass


class ChinupQueue(object):
    """
    List of pending Chinups with a common app token.
    """
    def __init__(self, app_token):
        self.app_token = app_token
        self.chinups = []

    def append(self, req):
        self.chinups.append(req)

    def sync(self, caller=None):
        """
        Builds and sends batch request, then populates Chinup responses.
        """
        if caller:
            assert caller in self.chinups

        # Take the existing queue from self.chinups. This is the max we will
        # try to accomplish in this sync, even if more are added during
        # processing (this can happen in chinup callback, or for paged
        # responses).
        chinups, self.chinups = self.chinups, []

        # Some requests in the batch might time out rather than completing.
        # Continue batching until the calling chinup is satisfied, or until we
        # stop making progress.
        progress = 1
        chinups = [cu for cu in chinups if not cu.completed]

        while chinups and progress and not (caller and caller.completed):

            # Ask the first chinup to process the chinups into a list of
            # request dicts. This is a classmethod, but calling via the first
            # chinup doesn't require us to know if Chinup has been subclassed.
            chinups, requests = chinups[0].prepare_batch(chinups)

            # It's possible that prepare_batch() decided all the chinups
            # were invalid, so make sure that we actually have requests.
            if not requests:
                assert not chinups
                logger.debug("No requests in batch after calling make_request_dicts()")
                break

            # Make the batch request.
            assert len(requests) <= 50
            logger.log(logging.INFO if settings.DEBUG_REQUESTS else logging.DEBUG,
                       "Making batch request len=%s/%s queue=%s",
                       len(requests), len(chinups), id(self))
            responses = batch_request(self.app_token, requests)

            # Populate responses into chinups.
            for cu, r in zip(chinups, responses):
                # Don't set response for timeouts, so they'll be automatically
                # tried again when .data is accessed.
                if r is not None:
                    cu.response = r
                logger.log(logging.INFO if settings.DEBUG_REQUESTS else logging.DEBUG,
                           '%s%r', 'TIMEOUT ' if r is None else '', cu)

            # Check for progress.
            progress = sum(1 for cu in chinups if cu.completed)

            # Filter out the completed chinups for the next pass.
            chinups = [cu for cu in chinups if not cu.completed]

        if caller and not caller.completed:
            # Ugh, this means we timed out without making progress.
            caller.exception = QueueTimedOut("Couldn't make enough progress to complete request.")

        # Drop completed chinups from the queue to prevent clogging with
        # completed chinups. Put them on the front of the queue, rather than
        # replacing it entirely, in case there were callbacks (in the response
        # setter) that added to self.chinups.
        self.chinups[:0] = chinups

    def __getstate__(self):
        raise AssertionError("Why are you pickling a ChinupQueue?")



from __future__ import absolute_import, unicode_literals

from .lowlevel import batches
from .queue import delete_queues
from . import settings


class ChinupTestMixin(object):
    """
    Clear batches record in setUp, and provide assertBatches.
    """
    def setUp(self):
        super(ChinupTestMixin, self).setUp()
        settings.TESTING = True
        delete_queues()
        batches[:] = []

    def assertBatches(self, nb, nr):
        lb, lr = len(batches), sum(len(b) for b in batches)
        if (lb, lr) != (nb, nr):
            from pprint import pformat
            pretty_batches = pformat(
                [['{} {}'.format(r['method'], re.sub(r'\?.*', '', r['relative_url']))
                  for r in b] for b in batches])
            raise AssertionError("Batches/requests actual {}/{} != expected {}/{}:\n{}".format(
                lb, lr, nb, nr, pretty_batches))

        # Reset for the next call
        batches[:] = []



from __future__ import absolute_import, unicode_literals

from decimal import Decimal
import hashlib
import imghdr
import json
import logging
import mimetypes
import os
import sys

import requests
from requests.utils import guess_filename

from . import settings
from .cache import get_cache
from .exceptions import OAuthError, TransportError, FacebookError


logger = logging.getLogger(__name__)


batches = []

def batch_request(app_token, reqs, url='https://graph.facebook.com'):
    """
    Runs a batch request to the Facebook API.
    """
    required_keys = set(['method', 'relative_url'])
    assert all(required_keys <= set(r) for r in reqs)

    # Split out binary attachments.
    # https://developers.facebook.com/docs/graph-api/making-multiple-requests/#binary
    #
    # The batch API doesn't provide any way to name the file upload fields.
    # For example, normal upload to me/photos uses the "source" parameter
    # to include a file; batch upload to me/photos specifies
    # "attached_files."  We hope this also works for other upload endpoints
    # such as ad creation.  It will break down for any endpoint that
    # requires uploading more than one file simultaneously.
    files = {}
    for i, req in enumerate(reqs, 1):
        if 'files' in req:
            reqfiles = {'file{}-{}'.format(i, j): f
                        for j, f in enumerate(req['files'].values())}
            assert all(f not in files for f in reqfiles)
            files.update(reqfiles)
            req['attached_files'] = ','.join(reqfiles)
            del req['files']

    # Upgrade files to tuples with content-type.
    files = {k: file_tuple(f, k) for k, f in files.items()}

    # Add etags headers.
    if settings.ETAGS:
        edicts = add_etags(reqs, app_token)

    # Similar to django.db.connection.queries, save batched requests in
    # debug mode for inspection.
    if settings.DEBUG or settings.TESTING:
        batches.append(reqs)

    # Post the batch request. Always include headers, for
    #  1. 302 location header on biz picture image,
    #  2. etags support,
    #  3. debugging.
    data = dict(access_token=app_token,
                batch=json.dumps(reqs),
                include_headers='true')
    try:
        r = requests.post(url, data=data, files=files)
    except requests.RequestException as e:
        raise TransportError(e)

    # Attempt to parse before checking HTTP status, because
    # parse_fb_response() will raise an exception for Facebook enumerated
    # error responses.
    resps = parse_fb_response(r)
    if r.status_code != 200:
        raise FacebookError("HTTP {}: {}".format(r.status_code, resps))
    if not isinstance(resps, list):
        raise FacebookError(resps)

    # Handle etags in responses.
    if settings.ETAGS:
        resps = handle_etags(resps, edicts)

    # Check for a timeout, unambiguously represented by null in the JSON,
    # or None when decoded.
    # https://developers.facebook.com/docs/graph-api/making-multiple-requests/#timeouts
    timed_out = sum(1 for r in resps if r is None)
    if timed_out:
        logger.warning("Timed out %d in batch of %d requests.", timed_out, len(reqs))

    return resps


def parse_fb_response(response):
    data = response.content.decode('utf-8')
    try:
        data = json.loads(data, parse_float=Decimal)
    except ValueError:
        pass
    else:
        exc = parse_fb_exception(data)
        if exc:
            raise exc
    return data


def parse_fb_exception(data):
    if data is False:
        return FacebookError("Facebook returned false")

    if isinstance(data, dict) and 'error' in data:
        error = data['error']
        if error.get('type') == 'OAuthException':
            eclass = OAuthError
        else:
            eclass = FacebookError
        return eclass(error.get('message'), error.get('code'))


def file_tuple(f, default_name=None, image=True):
    """
    Returns a tuple suitable for requests file upload, particularly to enable
    Facebook APIs that require the content-type to be set.
    """
    # Same variable names as in RequestEncodingMixin._encode_files().
    fn, fp, ft, fh = None, None, None, None

    if isinstance(f, (tuple, list)):
        # Assign as many as provided.
        try:
            fn = f[0]
            fp = f[1]
            ft = f[2]
            fh = f[3]
        except IndexError:
            pass
    else:
        # As in requests, bare value is file pointer or content.
        fp = f

    if not fn:
        fn = guess_filename(fp) or default_name

    # Try to get content-type from filename first.
    if not ft and fn:
        ext = os.path.splitext(fn)[1].lower()
        ft = mimetypes.types_map.get(ext)

    # Fall back to detecting the image type from the header.
    if not ft and image:
        ext = None
        if isinstance(fp, basestring):
            ext = imghdr.what(None, fp[:32])
        elif callable(getattr(fp, 'seek', None)):
            try:
                ext = imghdr.what(fp)
            except IOError:  # unseekable file
                pass
        if ext:
            ft = mimetypes.types_map.get(ext)

    return fn, fp, ft, fh


def add_etags(requests, app_token):
    """
    Returns a list of dicts like this:

      edicts = [ { request, key, responses },
                 { request, key, responses },
               ]

    where responses is an entry from cache with this structure:

      [ (etag, response),
        (etag, response),
      ]

    The list of dicts (edicts) is passed back to caller, so it can be reused by
    handle_etags below.
    """
    cache = get_cache()
    edicts = [dict(request=r, key=etags_cache_key(r, app_token))
              for r in requests]
    if cache:
        responses = cache.get_many([e['key'] for e in edicts])
    else:
        logger.warning("Chinup ETAGS=True but CACHE=%r", cache)
        responses = {}
    edicts = [dict(e, responses=responses.get(e['key'], []))
              for e in edicts]

    for edict in edicts:
        if edict['responses']:
            headers = edict['request'].setdefault('headers', [])
            assert all(not h.lower().startswith('if-none-match:')
                       for h in headers)
            headers.append("If-None-Match: {}".format(
                ', '.join(e for e, r in edict['responses'])))

    return edicts


def handle_etags(responses, edicts):
    """
    Restores cached responses, and saves new etagged responses to cache.
    Returns the new list of responses.
    """
    # We must have at least as many edicts as responses, otherwise
    # new_responses will be truncated.
    assert len(edicts) >= len(responses)

    new_responses = []
    to_cache = {}

    for response, edict in zip(responses, edicts):
        if isinstance(response, dict) and 'headers' in response:
            # Check for an etag in the response (whether 304 or 200).
            etag = next((h['value'] for h in response['headers']
                         if h['name'].lower() == 'etag'), None)

            # Check for a 304 response, look for matching etag.
            if response['code'] == 304:
                if etag:
                    assert edict['responses']
                    response = next(r for e, r in edict['responses'] if e == etag)
                else:
                    assert len(edict['responses']) == 1
                    etag, response = edict['responses'][0]

                logger.debug("Got 304 etag=%s, replacing with %s",
                             etag, response['code'])

            # Promote this etag to front of cache list for this request.
            if etag:
                resps = [(etag, response)]
                resps.extend((e, r) for e, r in edict['responses'] if e != etag)
                # Facebook should return an ETag header with a 304
                # response, but it does not. This means we can't actually cache
                # more than one response per request at a time, which is
                # annoying but probably fine.
                resps = resps[:1]
                to_cache[edict['key']] = resps

        new_responses.append(response)

    # Update cached etags.
    cache = get_cache()
    if cache:
        cache.set_many(to_cache, timeout=86400)  # one day
    else:
        logger.warning("Chinup ETAGS=True but CACHE=%r", cache)

    assert len(new_responses) == len(responses)
    return new_responses


def etags_cache_key(request, app_token):
    assert all(isinstance(k, basestring) for k in request.keys())
    assert all(isinstance(v, basestring) for v in request.values())
    m = hashlib.md5(repr(sorted(request.items())))
    m.update(app_token)
    key = 'chinup.etags.' + m.hexdigest()
    assert len(key) <= 250
    return key



from __future__ import absolute_import, unicode_literals


APP_TOKEN = None
DEBUG = False
DEBUG_REQUESTS = DEBUG
DEBUG_HEADERS = False
TESTING = False
ETAGS = True
CACHE = None


try:
    from django.conf import settings
    for name in dir(settings):
        if name.startswith('CHINUP_'):
            locals()[name[7:]] = getattr(settings, name)
except ImportError:
    pass


__all__ = [name for name in locals().keys() if name.isupper()]



from __future__ import absolute_import, unicode_literals


class ChinupError(Exception):
    """Base class for exceptions raised by chinup."""


class FacebookError(ChinupError):
    """Exception for Facebook enumerated errors."""
    def __init__(self, message=None, code=None):
        self.code = code
        if code is not None:
            message = '[{}] {}'.format(code, message)
        super(FacebookError, self).__init__(message)


class OAuthError(FacebookError):
    """Exception for Facebook errors specifically related to OAuth."""


class TransportError(ChinupError):
    """Exception for transport errors."""
    def __init__(self, message=None):
        self.e = None
        if isinstance(message, Exception):
            self.e = message
            message = '{}: {}'.format(message.__class__.__name__, message)
        super(TransportError, self).__init__(message)


class QueueTimedOut(ChinupError):
    pass


class PagingError(ChinupError):
    pass


class ChinupCanceled(ChinupError):
    pass



from __future__ import absolute_import, unicode_literals

try:
    from .allauth import *
except ImportError:
    from .chinup import *

from .exceptions import *


__version__ = '0.2.1'


# Configure logging to avoid warning.
# https://docs.python.org/2/howto/logging.html#configuring-logging-for-a-library
import logging
if hasattr(logging, 'NullHandler'):
    logging.getLogger('chinup').addHandler(logging.NullHandler())



from __future__ import absolute_import, unicode_literals

import sys

from . import settings


def get_cache():
    if not hasattr(settings, '_CACHE'):
        settings._CACHE = None
        if isinstance(settings.CACHE, basestring):
            pkg, attr = settings.CACHE.rsplit('.', 1)
            __import__(pkg)
            settings._CACHE = getattr(sys.modules[pkg], attr)
    return settings._CACHE
